{
  @exclude

  @html(<b>)
  StreamSec SSL Plug-in
  @html(</b>)
  - Copyright (c) Henrick Hellstrom & Danijel Tkalcec
  @html(<br><br>)

  This unit defines Plug-ins for using StreamSec 2.1 SSL components 
  with rtcHttpClient and rtcHttpServer for SSL encryption/decryption.
}
unit rtcSSecPlugin;

interface

uses
  Classes, SysUtils,

  // RTC SDK
  rtcPlugins,

  // StreamSec Tools 2.1
  TlsClass,
  StreamSecII,
  TlsInternalServer;

type
  // Basic StreamSec Plug-in implementation
  TRtcSSecPlugin = class(TRtcCryptPlugin)
  private
    FTLSServer: TCustomTLSInternalServer;
    FBuffer: string;
    procedure SetTLSServer(const Value: TCustomTLSInternalServer);
  protected
    procedure CheckTLSServer; virtual; abstract;
    procedure Notification(AComponent: TComponent; Operation: TOperation);
      override;
  public
    { Called before we do a graceful disconnect, in case some data has to be sent out. }
    procedure BeforeDisconnect(ID:TRtcConnID; var OutData:string); override;

    { Called after a connection was closed. }
    procedure AfterDisconnect(ID:TRtcConnID); override;

    { Called when data arrived.
      InData = data received from recipient (decode that!)
      OutData = data prepared by "decoder" for sending back to recipient (encoded data)
      OutPlainText = decrypted input data (for use by application) }
    procedure DataReceived(ID:TRtcConnID; const InData:string; var OutData:string; var OutPlainText:string); override;

    { Called when data needs to be sent.
      InData = application data which we want to be encoded for sending
      OutData = encoded data which should be sent out }
    procedure DataToSend(ID:TRtcConnID; const InData:string; var OutData:string); override;
  published
    property TLSServer: TCustomTLSInternalServer read FTLSServer write SetTLSServer;
    end;

  { StreamSec Plugin for Server-side SSL encryption/decryption }
  TRtcSSecServerPlugin = class(TRtcSSecPlugin)
  protected
    procedure CheckTLSServer; override;
  public
    { Called after a new connection was established.
      OutData = data which has to be sent out immediately. }
    procedure AfterConnect(ID:TRtcConnID; var OutData:string); override;
  end;

  { StreamSec Plugin for client-side SSL encryption/decryption }
  TRtcSSecClientPlugin = class(TRtcSSecPlugin)
  protected
    procedure CheckTLSServer; override;
  public
    { Called after a new connection was established.
      OutData = data which has to be sent out immediately. }
    procedure AfterConnect(ID:TRtcConnID; var OutData:string); override;
  end;

implementation

function RtcConnIDToPointer(ID: TRtcConnID): Pointer;
begin
  Result := Pointer(ID);
end;

{ TRtcSSecPlugin }

procedure TRtcSSecPlugin.AfterDisconnect(ID: TRtcConnID);
var
  lDummy: string;
begin
  BeforeDisconnect(ID,lDummy);
end;

procedure TRtcSSecPlugin.BeforeDisconnect(ID: TRtcConnID; var OutData: string);
var
  lRes: TStringStream;
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  lRes := TStringStream.Create('');
  try
    FTLSServer.TLSClose(RtcConnIDToPointer(ID),lRes);
    OutData := lRes.DataString;
  finally
    lRes.Free;
  end;
end;

procedure TRtcSSecPlugin.DataReceived(ID: TRtcConnID; const InData: string;
  var OutData, OutPlainText: string);
var
  lIn, lOut, lPT: TStringStream; 
  Client: TCustomTLS_ContentLayer;
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  if FTLSServer.FindTLSSession(RtcConnIDToPointer(ID),Client) < 0 then
    Client := nil;
  lIn := TStringStream.Create(InData);
  lOut := TStringStream.Create('');
  lPT := TStringStream.Create('');
  try
    while lIn.Size>lIn.Position do
      if Assigned(Client) then
        Client.DecodeData(lIn,lPT,lOut)
      else
        FTLSServer.TLSDecodeData(RtcConnIDToPointer(ID),lIn,lPT,lOut);

    if (FBuffer <> '') and Assigned(Client) and Client.Encrypted then begin
      lIn.Size := 0;
      lIn.WriteString(FBuffer);
      lIn.Position := 0;
      FBuffer := '';
      while lIn.Size>lIn.Position do
        Client.EncodeData(lIn,lOut);
    end;
    OutData := lOut.DataString;
    OutPlainText := lPT.DataString;
  finally
    lIn.Free;
    lOut.Free;
    lPT.Free;
  end;
end;

procedure TRtcSSecPlugin.DataToSend(ID: TRtcConnID; const InData: string;
  var OutData: string);
var
  lIn, lOut: TStringStream;
  Client: TCustomTLS_ContentLayer;
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  if FTLSServer.FindTLSSession(RtcConnIDToPointer(ID),Client) < 0 then
    raise Exception.Create('Connection is closed');
  if Client.Encrypted then begin
    lOut := TStringStream.Create('');
    lIn := TStringStream.Create(FBuffer + InData);
    try
      FBuffer := '';
      while lIn.Size > lIn.Position do
        Client.EncodeData(lIn,lOut);

      OutData := lOut.DataString;
    finally
      lIn.Free;
      lOut.Free;
    end;
  end else begin
    FBuffer := FBuffer + InData;
    OutData := '';
  end;
end;

procedure TRtcSSecPlugin.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited;
  if Operation = opRemove then begin
    if AComponent = FTLSServer then
      FTLSServer := nil;
  end;
end;

procedure TRtcSSecPlugin.SetTLSServer(const Value: TCustomTLSInternalServer);
begin
  if Assigned(FTLSServer) then
    FTLSServer.RemoveFreeNotification(Self);
  FTLSServer := Value;
  CheckTLSServer;
  if Assigned(FTLSServer) then
    FTLSServer.FreeNotification(Self);
end;

{ TRtcSSecServerPlugin }

type
  TTLSServerHack = class(TCustomTLSInternalServer);

procedure TRtcSSecServerPlugin.AfterConnect(ID: TRtcConnID; var OutData: string);
var
  Client: TCustomTLS_ContentLayer;
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  OutData := '';
  if FTLSServer.FindTLSSession(RtcConnIDToPointer(ID),Client) >= 0 then
    raise Exception.Create(Format('ID %d already in use',[ID]));
end;

procedure TRtcSSecServerPlugin.CheckTLSServer;
begin
  if Assigned(FTLSServer) then
    if FTLSServer.ClientOrServer = cosClientSide then begin
      FTLSServer := nil;
      raise Exception.Create('ClientOrServer must be set to cosServerSide');
    end;
end;

{ TRtcSSecClientPlugin }

procedure TRtcSSecClientPlugin.AfterConnect(ID: TRtcConnID; var OutData: string);
var
  lRes: TStringStream;
begin
  Assert(Assigned(FTLSServer),'TLSServer must be assigned');
  lRes := TStringStream.Create('');
  try
    FTLSServer.TLSConnect(RtcConnIDToPointer(ID),lRes);
    OutData := lRes.DataString;
  finally
    lRes.Free;
  end;
end;

procedure TRtcSSecClientPlugin.CheckTLSServer;
begin
  if Assigned(FTLSServer) then
    if FTLSServer.ClientOrServer = cosServerSide then begin
      FTLSServer := nil;
      raise Exception.Create('ClientOrServer must be set to cosClientSide');
    end;
end;

end.


